/* Frequent Geometric Subgraph Retrieval (FreqGeo).
 *
 * Copyright (C) 2008 -- Sebastian Nowozin <sebastian.nowozin@tuebingen.mpg.de>
 * This file is part of FreqGeo.
 *
 * FreqGeo is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, version 3 of the License.
 *
 * FreqGeo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with FreqGeo.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <vector>
#include <map>
#include <utility>
#include <algorithm>

#include <FreqGeo.h>

namespace FriGeo
{

void
FreqGeo::Report (const Graph& subg,
	const std::map<unsigned int, std::vector<std::vector<unsigned int> > >& L)
{
	std::cout << std::endl;
	std::cout << "=======================================================" << std::endl;
	std::cout << "FreqGeo::Report, support " << L.size () << ", in: ";
	for (std::map<unsigned int, std::vector<std::vector<unsigned int> > >::const_iterator
		li = L.begin () ; li != L.end () ; ++li)
	{
		std::cout << li->first << " ";
	}
	std::cout << std::endl;

	std::cout << "GRAPH" << std::endl << subg << std::endl;
}

bool
FreqGeo::IsFirstTimeVisitSubgraph (const Graph& subg,
	const std::vector<unsigned int>& PI,
	const std::map<unsigned int, std::vector<std::vector<unsigned int> > >& L) const
{
	// In case we cannot track the reference graph location list, so always
	// return true
	if (refg_index < 0)
		return (true);

	// Get L[refg]
	std::map<unsigned int, std::vector<std::vector<unsigned int> > >::const_iterator
		refg_L = L.find(refg_index);
	assert(refg_L != L.end());
	assert(refg_L->second.size() >= 1);
	unsigned int count_in_refg = refg_L->second.size();

	if (count_in_refg <= 1)
		return (true);	// subgraph appears only once, hence is unique

	// Subgraph appears more than once, check if its the first occurence
	const std::vector<unsigned int>& refg_L_first = refg_L->second[0];
	assert(refg_L_first.size() == PI.size());
	bool is_first = std::equal(refg_L_first.begin(), refg_L_first.end(),
		PI.begin());

	return (is_first);
}

void
FreqGeo::MineR (const Graph& subg, const std::vector<unsigned int>& PI,
	const std::map<unsigned int, std::vector<std::vector<unsigned int> > >& L)
{
	/* 1. Test minimum support
	 */
	if (L.size () < minsup)
		return;

	/* 2. Report subg if and only if:
	 *     i) subg appears only once in refg, or
	 *    ii) it appears multiple times but subg is the first in the location
	 *        list L[refg]
	 * Even if the subgraph is not reported, we need to further track its
	 * extensions
	 */
	if (IsFirstTimeVisitSubgraph (subg, PI, L)) {
		graphCount += 1;
		Report (subg, L);
	}

	/* 3. Obtain extensions of subg in refg.
	 */
	std::vector<Graph> extensionsEdge;
	std::vector<std::pair<Graph, std::vector<unsigned int> > >
		extensionsEdgeVertex;
	refg.ListExtensions (subg, PI, extensionsEdge, extensionsEdgeVertex);

#ifdef DEBUG
	std::cout << extensionsEdge.size () << " edge extensions, "
		<< extensionsEdgeVertex.size () << " edge-vertex extensions." << std::endl;
#endif

	/* 4. For each edge-vertex extension found, recurse and perform
	 *    projections of the location list.  We perform edge-vertex extensions
	 *    first in order to first cover all spanning trees of refg.
	 */
	for (std::vector<std::pair<Graph, std::vector<unsigned int> > >::iterator
		evi = extensionsEdgeVertex.begin () ; evi != extensionsEdgeVertex.end () ;
		++evi)
	{
		const Graph& eg = evi->first;
		const std::vector<unsigned int>& egPI = evi->second;

		/* The location list preserved under this extension
		 */
		std::map<unsigned int, std::vector<std::vector<unsigned int> > > Lsub;

		/* For each graph in the location list, try to extend
		 */
		for (std::map<unsigned int, std::vector<std::vector<unsigned int> > >::const_iterator
			li = L.begin () ; li != L.end () ; ++li)
		{
#ifdef DEBUG
			std::cout << "  In database graph " << li->first << " we have "
				<< li->second.size () << " locations" << std::endl;
#endif
			/* For each match in the current graph's location list, try to
			 * extend this specific match
			 */
			for (std::vector<std::vector<unsigned int> >::const_iterator
				lis = li->second.begin () ; lis != li->second.end () ; ++lis)
			{
				std::vector<std::vector<unsigned int> > matches;

				const Graph& curG = G[li->first];
				bool doesMatch = curG.Extend (subg, *lis,
					eg.LastVertexGraph (), eg.LastEdge (),
					matches, vertexTolerance);

				assert (doesMatch == false || matches.size () > 0);

#ifdef DEBUG
				std::cout << "  matches: " << (doesMatch ? "yes" : "no") << std::endl;
				std::cout << "      Lsub.size () = " << Lsub.size () << std::endl;
#endif

				if (doesMatch == false)
					continue;

				/* Add matches into new location list
				 * TODO: this should be done better using STL insert
				 */
				for (std::vector<std::vector<unsigned int> >::const_iterator
					mi = matches.begin () ; mi != matches.end () ; ++mi)
				{
					Lsub[li->first].push_back (*mi);
				}
			}
		}

		if (Lsub.size () > 0) {
#ifdef DEBUG
			std::cout << "Found " << Lsub.size ()
				<< " edge-vertex extensions, recursing" << std::endl;
#endif

			MineR (eg, egPI, Lsub);
		}
	}

	/* 5. For each edge-only extension, recurse and project.
	 */
	for (std::vector<Graph>::iterator evi = extensionsEdge.begin () ;
		evi != extensionsEdge.end () ; ++evi)
	{
		Graph emptyV;
		emptyV.Seal ();

		std::map<unsigned int, std::vector<std::vector<unsigned int> > > Lsub;

		for (std::map<unsigned int, std::vector<std::vector<unsigned int> > >::const_iterator
			li = L.begin () ; li != L.end () ; ++li)
		{
			for (std::vector<std::vector<unsigned int> >::const_iterator
				lis = li->second.begin () ; lis != li->second.end () ; ++lis)
			{
				std::vector<std::vector<unsigned int> > matches;
				const Graph& curG = G[li->first];
				bool doesMatch = curG.Extend (subg, *lis,
					emptyV, evi->LastEdge (), matches, vertexTolerance);

				if (doesMatch == false)
					continue;

				/* Add matches into new location list
				 * TODO: this should be done better using STL insert
				 */
				for (std::vector<std::vector<unsigned int> >::const_iterator
					mi = matches.begin () ; mi != matches.end () ; ++mi)
				{
					Lsub[li->first].push_back (*mi);
				}
			}
		}

#ifdef DEBUG
		std::cout << "Found " << Lsub.size ()
			<< " edge-only extensions, recursing" << std::endl;
#endif

		MineR (*evi, PI, Lsub);
	}
}

void
FreqGeo::Mine (void)
{
	Graph subg;
	subg.Seal ();
	std::vector<unsigned int> PI;

	/* As the empty graph is a subgraph of any graph, we initialize the
	 * location list with empty index mappings.
	 */
	std::map<unsigned int, std::vector<std::vector<unsigned int> > > L;
	for (unsigned int n = 0 ; n < G.size () ; ++n) {
		L[n] = std::vector<std::vector<unsigned int> > ();
		L[n].push_back (std::vector<unsigned int> ());
	}

	MineR (subg, PI, L);

	std::cout << "Visited " << graphCount << " subgraphs." << std::endl;
}

}

