/* Graph routines (FreqGeo).
 *
 * Copyright (C) 2008 -- Sebastian Nowozin <sebastian.nowozin@tuebingen.mpg.de>
 * This file is part of FreqGeo.
 *
 * FreqGeo is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, version 3 of the License.
 *
 * FreqGeo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with FreqGeo.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <algorithm>

#include <assert.h>

#include <Graph.h>
#include <Alignment.h>

namespace FriGeo
{

unsigned int
Graph::globalIdEdge (unsigned int edgeIndex) const
{
	return (VL.size () + edgeIndex);
}

unsigned int
Graph::globalIdVertex (unsigned int vertexIndex) const
{
	return (vertexIndex);
}

unsigned int
Graph::PushVertex (std::vector<double> pos, unsigned int lambda_v)
{
	assert (sealed == false);

	unsigned int row = gmm::mat_nrows (VP);
	gmm::resize (VP, row + 1, 4);

	gmm::copy (gmm::row_vector (pos), gmm::sub_matrix (VP,
		gmm::sub_interval (row, 1), gmm::sub_interval (0, 3)));
	VP(row,3) = 1;

	VL.push_back (lambda_v);

	return (row);
}

void
Graph::PushEdge (unsigned int v1, unsigned int v2, unsigned int lambda_e)
{
	assert (sealed == false);

	unsigned int row = gmm::mat_nrows (E);
	gmm::resize (E, row + 1, 4);

	/* Undirected edges, pre-sort.
	 */
	if (v1 > v2) {
		unsigned int stemp = v1;
		v1 = v2;
		v2 = stemp;
	}

	E(row,0) = v1;
	E(row,1) = v2;
	E(row,2) = lambda_e;
}

void
Graph::Seal (void)
{
	sealed = true;

	/* Build fast hash-array into edge list.
	 */
	edgemap.clear ();
	for (unsigned int er = 0 ; er < gmm::mat_nrows (E) ; ++er) {
		edgemap[E(er,0)].insert (er);
		edgemap[E(er,1)].insert (er);
	}
}


void
Graph::Unseal (void)
{
	sealed = false;
}


bool
Graph::Extend (const Graph& P, const std::vector<unsigned int> PI,
	const Graph& Q, const std::vector<unsigned int>& Qedge,
	std::vector<std::vector<unsigned int> >& matches,
	double matchingTolerance) const
{
	bool hasNode = Q.VL.size () > 0;
	bool hasEdge = Qedge.size () > 0;
	matches.clear ();

	/* Special case: add a single vertex to an empty graph.
	 */
	if (hasEdge == false) {
		assert (PI.size () == 0);	// Graph must be empty.
		assert (Q.VL.size () == 1);

		/* Simple match all individual vertices by label
		 */
		for (unsigned int vidx = 0 ; vidx < VL.size () ; ++vidx) {
			if (VL[vidx] != Q.VL[0])
				continue;

			std::vector<unsigned int> curmatch;
			curmatch.push_back (vidx);
			matches.push_back (curmatch);
		}

		return (matches.size () > 0);
	}

	/* 1. Get list of possible extension edges from mapped origin node
	 */
	std::map<unsigned int, std::set<unsigned int> >::const_iterator
		ci = edgemap.find (PI[Qedge[0]]);

	/* No edges found.
	 */
	if (ci->first != PI[Qedge[0]])
		return (false);

	/* 2. For each extension edge, check label, then try to extend P in this
	 *    graph using optimal alignment.
	 */
	gmm::dense_matrix<double> PVP (gmm::mat_nrows (P.VP) + 1, 4);
	gmm::copy (P.VP, gmm::sub_matrix (PVP,
		gmm::sub_interval (0, gmm::mat_nrows (P.VP)),
		gmm::sub_interval (0, 4)));

	/* The new vertex
	 */
	gmm::copy (Q.VP, gmm::sub_matrix (PVP,
		gmm::sub_interval (gmm::mat_nrows (P.VP), 1),
		gmm::sub_interval (0, 4)));
	//std::cout << "PVP" << std::endl << PVP << std::endl;

	gmm::dense_matrix<double> GVP (gmm::mat_nrows (P.VP) + 1, 4);
	gmm::copy (gmm::sub_matrix (VP, gmm::sub_index (PI),
			gmm::sub_interval (0,4)),
		gmm::sub_matrix (GVP, gmm::sub_interval (0, gmm::mat_nrows (P.VP)),
			gmm::sub_interval (0, 4)));

	for (std::set<unsigned int>::const_iterator si = ci->second.begin () ;
		si != ci->second.end () ; ++si)
	{
#ifdef DEBUG
		std::cout << "CONSIDER edge " << *si << ", from "
			<< E(*si,0) << " to " << E(*si,1) << ", label "
			<< E(*si,2) << std::endl;
#endif

		if (E(*si,2) != Qedge[2])
			continue;	// label mismatch, skip edge

		/* Check if the edge is within-graph
		 */
		if (hasNode == false) {
			if ((E(*si,0) == PI[Qedge[0]] && E(*si,1) == PI[Qedge[1]]) ||
				(E(*si,1) == PI[Qedge[0]] && E(*si,0) == PI[Qedge[1]]))
			{
				/* Output match here and return, as only a single match is
				 * possible (all vertices are mapped already)
				 *
				 * Note that as we only add an edge and no vertex, the output
				 * match is the same as the input one.
				 */
				matches.push_back (PI);

				return (true);
			}

			/* No further check necessary: only vertices have coordinates
			 */
			continue;
		}

		/* Try to build transformation from P space (PVP) into this graph.
		 */
		unsigned int targetVertexIndex = E(*si,1);
		if (E(*si,1) == PI[Qedge[0]])	// reverse edge?
			targetVertexIndex = E(*si,0);

		/* But first check if this vertex is already mapped
		 */
		std::vector<unsigned int>::const_iterator fi =
			std::find (PI.begin (), PI.end (), targetVertexIndex);
		if (fi != PI.end ()) {
#ifdef DEBUG
			std::cout << "   already mapped" << std::endl;
#endif
			continue;
		}

		/* The edge is a true extension edge, but does the target vertex label
		 * match?
		 */
		if (VL[targetVertexIndex] != Q.VL[0])
			continue;

		/* E(*si,target) is the vertex index of the new vertex to be matched
		 * with the node in Q.
		 */
		gmm::copy (gmm::sub_matrix (VP, gmm::sub_interval (targetVertexIndex, 1),
			gmm::sub_interval (0, 4)),
			gmm::sub_matrix (GVP,
				gmm::sub_interval (gmm::mat_nrows (GVP)-1, 1),
				gmm::sub_interval (0, 4)));

		/* Perform alignment
		 */
		//std::cout << "GVP" << std::endl << GVP << std::endl;
		Alignment align (PVP, GVP);
#ifdef DEBUG
		std::cout << "PVP" << std::endl << PVP << std::endl;
		std::cout << "GVP" << std::endl << GVP << std::endl;
#endif
		std::vector<double> singleLosses (gmm::mat_nrows (GVP));
		double fval = align.FindLeastSquaresAlignment (&singleLosses);

#ifdef DEBUG
		std::cout << "   fval = " << fval << std::endl;
#endif
		if (fval > matchingTolerance)
			continue;

		/* Output match
		 */
#ifdef DEBUG
		std::cout << "   adding match." << std::endl;
#endif
		matches.push_back (PI);
		matches[matches.size () - 1].push_back (targetVertexIndex);
	}

#ifdef DEBUG
	std::cout << matches.size () <<
		(matches.size () == 1 ? " match." : " matches.") << std::endl;
#endif

	return (matches.size () > 0);
}


int
Graph::MaximumNonBridgeIndex (const Graph& P,
	const std::vector<unsigned int>& PI) const
{
	if (P.edgemap.size () == 0)
		return (-1);

	std::vector<unsigned int> bridges;
	P.ListBridges (bridges);

	/* Search for maximum non-bridge index
	 */
	int NBmaxIdx = -1;
	for (std::vector<unsigned int>::const_iterator piv = PI.begin () ;
		piv != PI.end () ; ++piv)
	{
		/* For all edges in P originating from piv,
		 */
		unsigned int sourceIdx = piv - PI.begin ();
		std::map<unsigned int, std::set<unsigned int> >::const_iterator
			ci = P.edgemap.find (sourceIdx);

		/* Find out if they are bridging or not.
		 */
		for (std::set<unsigned int>::const_iterator si = ci->second.begin () ;
			si != ci->second.end () ; ++si)
		{
			assert (gmm::mat_nrows (P.E) > 0);

			bool bridging = std::find (bridges.begin (), bridges.end (), *si)
				!= bridges.end ();
			if (bridging)
				continue;

			unsigned int targetIdx;
			if (P.E(*si,0) == sourceIdx)
				targetIdx = P.E(*si,1);
			else
				targetIdx = P.E(*si,0);

			/* We found a non-bridging edge, now find its index in the
			 * original graph (this)
			 */
			unsigned int PsourceIdx = PI[sourceIdx];
			unsigned int PtargetIdx = PI[targetIdx];
			std::map<unsigned int, std::set<unsigned int> >::const_iterator
				ciT = edgemap.find (PsourceIdx);

			int edgeIndex = -1;
			for (std::set<unsigned int>::const_iterator siT = ciT->second.begin () ;
				siT != ci->second.end () ; ++siT)
			{
				if ((E(*siT,0) == PsourceIdx && E(*siT,1) == PtargetIdx) ||
					(E(*siT,1) == PsourceIdx && E(*siT,0) == PtargetIdx))
				{
					edgeIndex = *si;

					break;
				}
			}
			assert (edgeIndex != -1);

			if (edgeIndex > NBmaxIdx)
				NBmaxIdx = edgeIndex;
		}
	}

	return (NBmaxIdx);
}

void
Graph::ListExtensions (const Graph& P, const std::vector<unsigned int>& PI,
	std::vector<Graph>& extensionsEdge,
	std::vector<std::pair<Graph, std::vector<unsigned int> > >& extensionsEdgeVertex) const
{
	extensionsEdge.clear ();
	extensionsEdgeVertex.clear ();

	/* 1. Special case: PI is empty.  Then consider all single vertices in
	 *    this graph as extensions.
	 */
	if (PI.size () == 0) {
		for (unsigned int targetVertexIndex = 0 ;
			targetVertexIndex < VL.size () ; ++targetVertexIndex)
		{
			Graph Q(P);
			Q.Unseal ();

			std::vector<double> Vpos (3);
			gmm::copy (gmm::sub_matrix (VP,
				gmm::sub_interval (targetVertexIndex, 1),
				gmm::sub_interval (0, 3)), gmm::row_vector (Vpos));

			Q.PushVertex (Vpos, VL[targetVertexIndex]);
			Q.Seal ();

			std::vector<unsigned int> QI(PI);
			QI.push_back (targetVertexIndex);

			extensionsEdgeVertex.push_back (
				std::pair<Graph, std::vector<unsigned int> > (Q, QI));
		}

		return;
	}

	/* 2. Obtain maximum non-bridge edge index and maximum mapped vertex index
	 */
	int NBmaxIdx = MaximumNonBridgeIndex (P, PI);

	std::vector<unsigned int>::const_iterator maxPI =
		std::max_element (PI.begin (), PI.end ());
	int VmaxIdx = -1;
	if (maxPI != PI.end ())
		VmaxIdx = *maxPI;

	/* 3. Enumerate all possible edge-only extensions
	 */
	for (std::vector<unsigned int>::const_iterator piv = PI.begin () ;
		piv != PI.end () ; ++piv)
	{
		unsigned int PoriginIndex = piv - PI.begin ();

		/* All edges adjacent to vertex *piv
		 */
		std::map<unsigned int, std::set<unsigned int> >::const_iterator
			ci = edgemap.find (*piv);
		if (ci == edgemap.end())	// no adjacent edges? (1 node graph)
			continue;

		for (std::set<unsigned int>::const_iterator si = ci->second.begin () ;
			si != ci->second.end () ; ++si)
		{
#ifdef DEBUG
			std::cout << "E is " << gmm::mat_nrows(E) << "x" << gmm::mat_ncols(E)
				<< std::endl;
			std::cout << "node " << PoriginIndex << ", edge *si = " << *si
				<< std::endl;
			std::cout << "CONSIDER (from node " << PoriginIndex << ") edge " << *si << ", from "
				<< E(*si,0) << " to " << E(*si,1) << ", label "
				<< E(*si,2) << std::endl;
#endif

			/* Check that this edge index can be a maximum extension edge
			 */
			if (NBmaxIdx != -1 && *si <= ((unsigned int) NBmaxIdx)) {
#ifdef DEBUG
				std::cout << "   no maximum extension edge by bound" << std::endl;
#endif

				continue;
			}

			unsigned int targetVertexIndex = E(*si,1);
			if (E(*si,1) == *piv)	// reverse edge?
				targetVertexIndex = E(*si,0);

			/* If the target vertex of this edge is already mapped in P but the
			 * edge is not, we found a possible candidate for extension.
			 *
			 * Check whether the edge constitues a valid extension.
			 */
			std::vector<unsigned int>::const_iterator fi =
				std::find (PI.begin (), PI.end (), targetVertexIndex);

			bool isMapped = fi != PI.end ();
			if (isMapped == false)
				continue;

			if (E(*si,0) != PoriginIndex) {
				/* All edges are stored twice in the edgemap, hence we can
				 * skip over half of them.
				 */
				continue;
			}

			unsigned int PtargetIndex = fi - PI.begin ();

			/* Find a possible edge in P between PoriginIndex and
			 * PtargetIndex
			 */
			std::map<unsigned int, std::set<unsigned int> >::const_iterator
				pedge = P.edgemap.find (PoriginIndex);

			bool alreadyExist = false;
			for (std::set<unsigned int>::const_iterator psi = pedge->second.begin () ;
				psi != pedge->second.end () ; ++psi)
			{
				if (P.E(*psi,0) == PtargetIndex || P.E(*psi,1) == PtargetIndex) {
#ifdef DEBUG
					std::cout << "   edge already exists" << std::endl;
#endif

					alreadyExist = true;
					break;
				}
			}
			if (alreadyExist)
				continue;

			/* This edge does not exist in P, so it can be added.  But first
			 * check if the edge is a rightmost extension, otherwise
			 * skip it.
			 */
			Graph Q(P);
			Q.Unseal ();
			Q.PushEdge (PoriginIndex, PtargetIndex, E(*si,2));	// add edge
			Q.Seal ();
			int QNBmaxIdx = MaximumNonBridgeIndex (Q, PI);	// no vertex added
			if (QNBmaxIdx == -1)
				std::cout << "ERROR" << std::endl << Q << std::endl;
			assert (QNBmaxIdx != -1);

			if (*si < ((unsigned int) QNBmaxIdx)) {
				/* A non ordered edge addition.  This can happen in graphs like
				 *
				 *   P: 1 --A-- 2       Q: 1 --A-- 2
				 *       \                  \     /
				 *        C                  C   B
				 *         \                  \ /
				 *          3                  3
				 *
				 * Here, in P there is no non-bridge edge.  When edge B is
				 * added to produce Q, then edge C becomes the maximum-index
				 * non-bridge edge.  Hence f(Q) = Q \ {C} != P and adding edge
				 * B to P is a non-ordered operation.
				 */
#ifdef DEBUG
				std::cout << "   edge is non-maximal addition" << std::endl;
#endif

				continue;
			}

			/* A valid edge-only extension has been found.
			 */
			extensionsEdge.push_back (Q);
		}
	}

	/* If the graph has redundant edges, we don't need to perform vertex leaf
	 * extensions.
	 */
	bool isTree = (gmm::mat_nrows (P.E) < P.VL.size ());
	if (isTree == false) {
#ifdef DEBUG
		std::cout << "Has redundant edges, skipping vertex-only extensions." << std::endl;
#endif

		return;
	}

	/* The graph is a tree and has only bridge edges.  Perform leaf vertex
	 * extensions as follows.
	 *
	 * For all leaf nodes, add connecting edge and node with highest
	 * then-leaf-index.  Note that the "then-leaf-index" depends on which edge
	 * is used to connect a vertex, as the adjacent ex-leaf vertex becomes a
	 * non-leaf.
	 */
	for (unsigned int vi = 0 ; vi < P.VL.size () ; ++vi) {
		bool viHasEdges = (P.edgemap.find (vi) != P.edgemap.end ());

		std::set<unsigned int> edgesEmpty;
		const std::set<unsigned int>& PviEdges =
			viHasEdges ? P.edgemap.find (vi)->second : edgesEmpty;

#ifdef DEBUG
		std::cout << "CONSIDER vertex " << vi << ", "
			<< PviEdges.size () << " edges in P" << std::endl;
#endif

		if (viHasEdges && PviEdges.size () > 1)
			continue;	// vi is a non-leaf node

		/* We know: vi is a leaf node.  Enumerate edges in this graph which
		 * are not yet in P.
		 */
		bool thisHasEdges = (edgemap.find (PI[vi]) != edgemap.end ());
		const std::set<unsigned int>& viEdges =
			thisHasEdges ? edgemap.find (PI[vi])->second : edgesEmpty;

		if (viEdges.size () == PviEdges.size ())
			continue;	// All edges are added already

		for (std::set<unsigned int>::const_iterator evi = viEdges.begin () ;
			evi != viEdges.end () ; ++evi)
		{
			/* Consider the edge E(*evi,.)
			 */
			unsigned int sourceVertex = E(*evi,0);
			unsigned int destVertex = E(*evi,1);
#if 0
			if (sourceVertex != PI[vi])
				continue;
#endif

			if (destVertex == PI[vi]) {
				unsigned int tempVertex = sourceVertex;

				sourceVertex = destVertex;
				destVertex = tempVertex;
			}
			assert (sourceVertex == PI[vi]);

			/* If the destVertex already exist in P, then we would add a cycle
			 * edge, which is bad.
			 */
			if (std::find (PI.begin (), PI.end (), destVertex) != PI.end ())
				continue;

			bool foundEdge = false;
			for (std::set<unsigned int>::const_iterator pvi = PviEdges.begin () ;
				foundEdge == false && pvi != PviEdges.end () ; ++pvi)
			{
				if ((PI[P.E(*pvi,0)] == sourceVertex && PI[P.E(*pvi,1)] == destVertex) ||
					(PI[P.E(*pvi,1)] == sourceVertex && PI[P.E(*pvi,0)] == destVertex))
					foundEdge = true;
			}

			if (foundEdge)
				continue;	// edge does already exist in P

			/* Ok, here is what we have now: a leaf vertex vi in P, an edge
			 * from vi to a new vertex.
			 *
			 * Only one thing remains to check: would the vertex we add be the
			 * maximum vertex-index node in the new graph?
			 *
			 * We first check a shortcut: if the new index is larger than any
			 * index so far, it will be a valid extension.  If it is not, we
			 * need to check more carefully.
			 */
			bool isValidVertexExtension =
				(destVertex > *std::max_element (PI.begin (), PI.end ()));

			/* Special case: if the new graph has only two nodes, both are
			 * leaf-vertices.
			 */
			if (isValidVertexExtension == false && P.VL.size () > 1) {
				/* More careful check: check against the maximum of all
				 * then-leaf nodes's indices.
				 */
				int maximumIndex = -1;
				for (unsigned int mvi = 0 ; mvi < PI.size () ; ++mvi) {
					if (mvi == vi)
						continue;	// vi becomes a non-leaf node.

					if (P.edgemap.find (mvi)->second.size () > 1)
						continue;	// no leaf node

					if (((int) PI[mvi]) > maximumIndex)
						maximumIndex = PI[mvi];
				}

				isValidVertexExtension = (((int) destVertex) > maximumIndex);
			}

			if (isValidVertexExtension == false)
				continue;

			/* Valid extension: from vi in P, by edge E(*evi,.) to a new
			 * vertex 'destVertex'
			 */
			Graph Q(P);
			Q.Unseal ();
			std::vector<double> Vpos (3);
			gmm::copy (gmm::sub_matrix (VP,
				gmm::sub_interval (destVertex, 1),
				gmm::sub_interval (0, 3)), gmm::row_vector (Vpos));

			unsigned int PnewVertexIndex = gmm::mat_nrows (Q.VP);
			Q.PushVertex (Vpos, VL[destVertex]);
			Q.PushEdge (vi, PnewVertexIndex, E(*evi,2));	// add edge
			Q.Seal ();

			std::vector<unsigned int> QI(PI);
			QI.push_back (destVertex);

			extensionsEdgeVertex.push_back (
				std::pair<Graph, std::vector<unsigned int> > (Q, QI));
		}
	}

}

void
Graph::ListBridges (std::vector<unsigned int>& bridges) const
{
	bridges.clear ();
	if (gmm::mat_nrows (VP) == 0 || gmm::mat_nrows (E) == 0)
		return;

	std::vector<int> dfsOrder (gmm::mat_nrows (VP), -1);

	unsigned int order = 0;
	ListBridgesR (-1, order, 0, dfsOrder, bridges);

#ifdef DEBUG
	for (unsigned int n = 0 ; n < bridges.size () ; ++n)
		std::cout << "BRIDGE from " << E(bridges[n],0) << " to "
			<< E(bridges[n],1) << std::endl;
#endif

	assert (order == gmm::mat_nrows (VP)-1);
}

unsigned int
Graph::ListBridgesR (int parent, unsigned int& order, unsigned int id,
	std::vector<int>& dfsOrder, std::vector<unsigned int>& bridges) const
{
#ifdef DEBUG
	std::cout << "ListBridgesR, order " << order << ", id " << id << std::endl;
#endif

	dfsOrder[id] = order;	// Reached vertex 'id'
	unsigned int minReachable = order;

	/* Recurse into all edges
	 */
	std::map<unsigned int, std::set<unsigned int> >::const_iterator
		ci = edgemap.find (id);

	for (std::set<unsigned int>::const_iterator si = ci->second.begin () ;
		si != ci->second.end () ; ++si)
	{
		unsigned int targetVertexIndex = E(*si,1);
		if (targetVertexIndex == id)	// reverse edge?
			targetVertexIndex = E(*si,0);

		/* We have an edge (id,targetVertexIndex) now.  Two possible cases:
		 *
		 *   i) We just came from there.  Skip.
		 *  ii) The target vertex is new.  Recurse.
		 * iii) The target vertex has been reached before.
		 */
		if (parent != -1 && targetVertexIndex == ((unsigned int) parent))
			continue;

		if (dfsOrder[targetVertexIndex] >= 0) {
			/* Backedge
			 */
			if (((unsigned int) dfsOrder[targetVertexIndex]) < minReachable) {
				minReachable = dfsOrder[targetVertexIndex];
#ifdef DEBUG
				std::cout << "   id " << id << " min reachable " << minReachable << std::endl;
#endif
			}
		} else {
			/* Recurse
			 */
			order += 1;
			unsigned int cOrder = order;
			unsigned int VminReachable = ListBridgesR (id, order,
				targetVertexIndex, dfsOrder, bridges);

			/* This edge is a bridge if and only if: VminReachable >= cOrder
			 */
			if (VminReachable >= cOrder)
				bridges.push_back (*si);

			/* Reachability is recursive, hence the minimum node is also
			 * reachable from the current one.
			 */
			if (VminReachable < minReachable) {
				minReachable = VminReachable;
#ifdef DEBUG
				std::cout << "   id " << id << " min reachable " << minReachable << std::endl;
#endif
			}
		}
	}

	return (minReachable);
}

Graph
Graph::LastVertexGraph (void) const
{
	Graph g;

	assert (gmm::mat_nrows (VP) > 0);
	std::vector<double> pos (3);
	gmm::copy (gmm::sub_matrix (VP, gmm::sub_interval (gmm::mat_nrows (VP) - 1, 1),
		gmm::sub_interval (0, 3)), gmm::row_vector (pos));
	g.PushVertex (pos, VL[VL.size () - 1]);
	g.Seal ();

	return (g);
}

std::vector<unsigned int>
Graph::LastEdge (void) const
{
	if (gmm::mat_nrows (E) == 0)
		return (std::vector<unsigned int> ());

	std::vector<unsigned int> edge (3);
	gmm::copy (gmm::sub_matrix (E, gmm::sub_interval (gmm::mat_nrows (E) - 1, 1),
		gmm::sub_interval (0, 3)), gmm::row_vector (edge));

	return (edge);
}

std::ostream&
operator<< (std::ostream& os, const FriGeo::Graph& g)
{
	if (g.VL.size () == 0) {
		os << "{empty}" << std::endl;

		return (os);
	}

	for (unsigned int vi = 0 ; vi < g.VL.size () ; ++vi) {
		os << vi << " " << g.VL[vi];
		os << " [" << g.VP(vi,0) << "," << g.VP(vi,1) << "," << g.VP(vi,2) << "]";

		if (g.edgemap.count (vi) > 0)
			os << " => (";

		bool first = true;
		for (std::map<unsigned int, std::set<unsigned int> >::const_iterator
			ci = g.edgemap.find (vi) ; ci != g.edgemap.end () && ci->first == vi ; ++ci)
		{
			for (std::set<unsigned int>::const_iterator si = ci->second.begin () ;
				si != ci->second.end () ; ++si)
			{
				if (first == false)
					os << ", ";
				first = false;

				os << g.E(*si,0) << "-" << g.E(*si,1) <<
					":" << g.E(*si,2);
			}
		}

		if (g.edgemap.count (vi) > 0)
			os << ")";
		os << std::endl;
	}

	return (os);
}

}

