
#ifndef	GRAPH_H
#define	GRAPH_H

/* Basic graph data structures.
 *
 * Author: Sebastian Nowozin <sebastian.nowozin@tuebingen.mpg.de>
 * Date: 17th August 2007
 */

#include <vector>
#include <map>
#include <set>
#include <iostream>

#include <gmm/gmm.h>


namespace FriGeo
{

class Graph
{
private:
	friend std::ostream& operator<< (std::ostream& os, const FriGeo::Graph& g);

	bool sealed;

	/* VP: (N,4) 3D vertex coordinates.
	 * VL: Vertex labels.
	 */
	gmm::dense_matrix<double> VP;
	std::vector<unsigned int> VL;

	/* (M,3) edge table: vertex, vertex, label rows.
	 */
	gmm::dense_matrix<unsigned int> E;

	/* edgemap[v1] is the set of edge indices in E.
	 */
	std::map<unsigned int, std::set<unsigned int> > edgemap;

	/* The two methods below define a total ordering for all vertices and
	 * edges for a single graph (this).
	 */
	unsigned int globalIdEdge (unsigned int edgeIndex) const;
	unsigned int globalIdVertex (unsigned int vertexIndex) const;

	/* Recursion for Tarjan's cutting edge algorithm.
	 *
	 * parent: The vertex id of the originating node (each call is an edge).
	 * order: The current DFS order.
	 * id: The current vertex considered.
	 * dfsOrder: The vertex order table.
	 * bridges: The output, storing all bridge edge indices.
	 *
	 * Return the minimum DFS order over the nodes reachable from id.
	 */
	unsigned int ListBridgesR (int parent, unsigned int& order, unsigned int id,
		std::vector<int>& dfsOrder, std::vector<unsigned int>& bridges) const;

	/* Among all non-bridge edges in P, obtain the corresponding maximum index
	 * in the current graph, determined by the mapping PI.
	 *
	 * Return -1 if no non-bridge edge in P exist, or the maximal edge index
	 * in the current graph otherwise.
	 *
	 * Note: This is an internal method used for uniquely enumerating
	 * subgraphs.
	 */
	int MaximumNonBridgeIndex (const Graph& P,
		const std::vector<unsigned int>& PI) const;

	/* Unseal a graph, invalidating its edgemap.
	 */
	void Unseal (void);

public:
	Graph (void)
		: sealed (false)
	{
	}

	const std::vector<unsigned int>& VertexLabel (void) const
	{
		return (VL);
	}

	const gmm::dense_matrix<double>& VertexPosition (void) const
	{
		return (VP);
	}

	const gmm::dense_matrix<unsigned int>& Edges (void) const
	{
		return (E);
	}

	/* Add a vertex to the graph.
	 *
	 * pos: 3-vector of 3D coordinates.
	 * lambda_v: The label of the vertex.
	 *
	 * Return the vertex index assigned to this vertex in the graph.
	 */
	unsigned int PushVertex (std::vector<double> pos, unsigned int lambda_v);

	/* Add an undirected edge to the graph.  The vertices must exist already.
	 *
	 * v1, v2: Vertex indices, with v1 != v2.
	 * lambda_e: The label of the edge.
	 */
	void PushEdge (unsigned int v1, unsigned int v2, unsigned int lambda_e);

	/* Seal graph, disallowing further modifications.
	 */
	void Seal (void);

	/* Extension check.
	 *
	 * P: Current matching graph.
	 * PI: Current index translation table, P(vi) is this(PI[vi]).
	 * Q: Single-node graph to add or empty if only an edge is added.
	 * Qedge: 3-vector of v1-v2-elabel to be added to P.  The direction is
	 *    always old-to-new, hence Qedge[0] is always already mapped, and
	 *    Qedge[1] may or may not be mapped.  If Qedge.size() == 0, then no
	 *    edge is added and this is used only to add the very first vertex.
	 *
	 * Return true if matches are possible, false otherwise.  All possible
	 * matches (size bounded by number of nodes of this graph) are listed in
	 * 'matches' as index translation table.
	 */
	bool Extend (const Graph& P, const std::vector<unsigned int> PI,
		const Graph& Q, const std::vector<unsigned int>& Qedge,
		std::vector<std::vector<unsigned int> >& matches,
		double matchingTolerance = 5e-2) const;

	/* List possible (rightmost) extensions of P (mapped by PI) in this graph.
	 *
	 * P: Current subgraph.
	 * PI: Current index translation table, P(vi) is this(PI[vi]).
	 * extensionsEdge: Edge-only extensions (graph Q, index mapping PI)
	 *    with exactly no vertex and one edge added to it.
	 * extensionsEdgeVertex: Edge-vertex extensions, (graph Q, index mapping
	 *    PI), with exactly one vertex and one edge added to it.
	 *
	 * To determine the order (rightmostness), the vertex and edge order in
	 * this graph are used.  Note that this implies that ListExtensions is
	 * only called on a single reference graph, never on the database graphs.
	 */
	void ListExtensions (const Graph& P, const std::vector<unsigned int>& PI,
		std::vector<Graph>& extensionsEdge,
		std::vector<std::pair<Graph, std::vector<unsigned int> > >& extensionsEdgeVertex) const;

	/* Use Tarjan's algorithm to list cutting edges (bridges) in a graph P in
	 * linear time O(V+E).
	 *
	 * P: The graph to find the bridges in.
	 * bridges: The indices into P.E of all bridges.
	 */
	void ListBridges (std::vector<unsigned int>& bridges) const;

	/* Helper function: return a single node graph of the last vertex.
	 */
	Graph LastVertexGraph (void) const;

	/* Helper function: return the last edge added.
	 */
	std::vector<unsigned int> LastEdge (void) const;
};

std::ostream& operator<< (std::ostream& os, const FriGeo::Graph& g);

};

#endif

