/* mexfreqgeo.cpp - MEX interface for frequent geometric subgraph mining.
 *
 * Copyright (C) 2008 -- Sebastian Nowozin <sebastian.nowozin@tuebingen.mpg.de>
 * This file is part of FreqGeo.
 *
 * FreqGeo is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, version 3 of the License.
 *
 * FreqGeo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with FreqGeo.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Date: 20th August 2007
 *
 * Usage:
 *
 * [graphs, count, occurence] = mexfreqgeo (G, refg, minsup, tolerance);
 *
 ***
 * Input
 *
 * G: (1,n) cellarray of n graph structures with this layout
 *    g.nodelabels: (n,1) discrete integer labels [L_1 ; L_2 ; ... ; L_n];
 *    g.nodepos: (n,3) real 3D coordinates of the vertex.
 *    g.edges: (m,2) edges, [from to] at each line:
 *       [e_1_{from} e_1_{to} edgelabel_1 ; ... ; e_m_{from} e_m_{to} edgelabel_m]
 *       The node indices go from 1 to n.
 * refg: Integer between 1 and n selecting the reference graph from the set G.
 *    Only frequent subgraphs of the reference graph are considered.
 * minsup: The minimum number of times a frequent subgraph has to appear in G.
 * tolerance: (optional) Vertex match tolerance.  Default is 1e-1.
 *
 ***
 * Output
 *
 * graphs: (1,p) cellarray of p frequent graph structures.  Structures as for G.
 * count: (1,p) integer frequency count of the found graphs.
 * occurence: (optional) (n,p) integer matrix encoding at (i,j) the number of
 *    times graph{j} matches into sample G{i}.
 */

#include <assert.h>
#include <string.h>
#include <stdlib.h>

#include <iostream>
#include <vector>
#include <map>

#include <mex.h>

#include <mexhelp.h>
#include <Graph.h>
#include <FreqGeo.h>

class
MexFreqGeo : public FriGeo::FreqGeo
{
private:
	std::vector<mxArray*> graphs;
	std::vector<unsigned int> count;
	std::vector<std::map<unsigned int, unsigned int> > occurence;

protected:
	virtual void Report (const FriGeo::Graph& subg,
		const std::map<unsigned int,
			std::vector<std::vector<unsigned int> > >& L)
	{
		// Find the occurence list of the reference graph.  This must exist
		// and be non-empty.
		std::map<unsigned int, std::vector<std::vector<unsigned int> > >::const_iterator
			refg_L = L.find(refg_index);
		assert(refg_L != L.end());
		assert(refg_L->second.size() >= 1);
		unsigned int count_in_refg = refg_L->second.size();

		std::cout << "MexFreqGeo::Report, size "
			<< subg.VertexLabel ().size ()
			<< ", appears in " << L.size() << " graph"
				<< (L.size() > 1 ? "s" : "")
			<< ", " << count_in_refg
				<< (count_in_refg > 1 ? " times" : " time")
				<< " in reference graph"
			<< std::endl;
		std::cout << subg << std::endl;

		/* Don't consider the empty graph
		 */
		if (subg.VertexLabel ().size () == 0)
			return;

		graphs.push_back (FriGeo::MexHelp::writeGraph (subg));
		count.push_back (L.size ());

		std::map<unsigned int, unsigned int> omap;
		for (std::map<unsigned int, std::vector<std::vector<unsigned int> > >::const_iterator
			li = L.begin () ; li != L.end () ; ++li)
		{
			omap[li->first] = li->second.size ();
		}
		occurence.push_back (omap);
	}

public:
	MexFreqGeo (const std::vector<FriGeo::Graph>& G,
		unsigned int refg_index, unsigned int minsup,
		double vertexTolerance)
		: FriGeo::FreqGeo (G, refg_index, minsup, vertexTolerance)
	{
	}

	/* [graphs, count, occurence]
	 *
	 * graphs: (1,p) cellarray of p frequent graph structures.  Structures as for G.
	 * count: (1,p) integer frequency count of the found graphs.
	 * occurence: (optional) (n,p) integer matrix encoding at (i,j) the number of
	 *    times graph{j} matches into sample G{i}.
	 */
	void
	writeMexLHS (unsigned int samplecount, int nlhs, mxArray* plhs[])
	{
		plhs[0] = mxCreateCellMatrix (1, graphs.size ());
		assert (plhs[0] != NULL);
		mxArray* graphsP = plhs[0];

		uint32_T* countP = NULL;
		if (nlhs >= 2) {
			plhs[1] = mxCreateNumericMatrix (1, graphs.size (), mxUINT32_CLASS, 0);
			assert (plhs[1] != NULL);
			countP = (uint32_T *) mxGetPr (plhs[1]);
		}

		uint32_T* occP = NULL;
		if (nlhs >= 3) {
			plhs[2] = mxCreateNumericMatrix (samplecount, graphs.size (),
				mxUINT32_CLASS, 0);
			assert (plhs[2] != NULL);
			occP = (uint32_T *) mxGetPr (plhs[2]);
		}

		for (unsigned int idx = 0 ; idx < graphs.size () ; ++idx) {
			mxSetCell (graphsP, idx, graphs[idx]);	// transfer ownership

			if (countP != NULL)
				countP[idx] = count[idx];

			if (occP != NULL) {
				for (std::map<unsigned int, unsigned int>::const_iterator
					li = occurence[idx].begin () ; li != occurence[idx].end () ; ++li)
				{
					occP[li->first + idx*samplecount] = li->second;
				}
			}
		}
	}

};


/* [graphs, count, occurence] = mexfreqgeo (G, refg, minsup, tolerance);
 */
void
mexFunction (int nlhs, mxArray* plhs[], int nrhs, const mxArray* prhs[])
{
	if (nrhs < 3 || nrhs > 4) {
		mexErrMsgTxt ("Wrong number of right hand side parameters (must be 3 or 4).\n");

		return;
	}

	if (nlhs < 2 || nlhs > 3) {
		mexErrMsgTxt ("Wrong number of left hand side parameters (must be 2 or 3).\n");

		return;
	}

	if (mxIsCell (prhs[0]) == 0) {
		mexErrMsgTxt ("G must be a cellarray of graph structures.\n");

		return;
	}

	if (mxIsNumeric (prhs[1]) == 0) {
		mexErrMsgTxt ("refg must be an integer.\n");
		return;
	}

	const mxArray* graphs = prhs[0];
	unsigned int gcount = mxGetN (graphs);
	unsigned int minsup = (unsigned int) mxGetScalar (prhs[2]);

	if (minsup > gcount) {
		mexErrMsgTxt ("The given minsup value is larger than the graph database size.\n");

		return;
	}

	const mxArray** garr = (const mxArray**) mxGetPr (graphs);
	std::vector<FriGeo::Graph> G;

	for (unsigned int n = 0 ; n < gcount ; ++n) {
		const mxArray* gcur = garr[n];

		FriGeo::Graph g;
		if (FriGeo::MexHelp::readGraph (gcur, g) == false)
			return;

		G.push_back (g);
	}

	unsigned int refg_index = static_cast<unsigned int>(mxGetScalar (prhs[1]));
	if (refg_index < 1 || refg_index > G.size()) {
		mexErrMsgTxt ("refg must be between 1 and n.\n");
		return;
	}

#if 0
	FriGeo::Graph refg;
	if (FriGeo::MexHelp::readGraph (garr[refg_index - 1], refg) == false)
		return;
#endif
#if 0
	if (FriGeo::MexHelp::readGraph (prhs[1], refg) == false)
		return;
#endif

	double tolerance = 1e-1;
	if (nrhs >= 4)
		tolerance = mxGetScalar (prhs[3]);

	if (tolerance < 0.0) {
		mexErrMsgTxt ("The given matching tolerance is negative.\n");

		return;
	}

	MexFreqGeo fg (G, refg_index - 1, minsup, tolerance);
	fg.Mine ();
	fg.writeMexLHS (G.size (), nlhs, plhs);
}

