/* mexisgeosubgraph.cpp - MEX interface for testing subgraph relationship.
 *
 * Copyright (C) 2008 -- Sebastian Nowozin <sebastian.nowozin@tuebingen.mpg.de>
 * This file is part of FreqGeo.
 *
 * FreqGeo is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, version 3 of the License.
 *
 * FreqGeo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with FreqGeo.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Date: 4th April 2008
 *
 * Usage:
 *
 * [count] = mexisgeosubgraph (G, subg, tolerance);
 *
 ***
 * Input
 *
 * G: (1,n) cellarray of n graph structures with this layout
 *    g.nodelabels: (n,1) discrete integer labels [L_1 ; L_2 ; ... ; L_n];
 *    g.nodepos: (n,3) real 3D coordinates of the vertex.
 *    g.edges: (m,2) edges, [from to] at each line:
 *       [e_1_{from} e_1_{to} edgelabel_1 ; ... ; e_m_{from} e_m_{to} edgelabel_m]
 *       The node indices go from 1 to n.
 * subg: Subgraph to test subgraph relationship with.  Same format as G.
 * tolerance: (optional) Vertex match tolerance.  Default is 1e-1.
 *
 ***
 * Output
 *
 * count: (n,1) integer vector encoding at i'th element the number of times
 *    subg appears in G{i}.
 */

#include <assert.h>
#include <string.h>
#include <stdlib.h>

#include <iostream>
#include <vector>
#include <map>
#include <algorithm>

#include <mex.h>

#include <mexhelp.h>
#include <Graph.h>
#include <FreqGeo.h>

class
MexIsGeoSubgraph : public FriGeo::FreqGeo
{
private:
	FriGeo::Graph subg_full;
	std::vector<unsigned int> count;

protected:
	virtual void Report (const FriGeo::Graph& subg,
		const std::map<unsigned int,
			std::vector<std::vector<unsigned int> > >& L)
	{
		// Check node size, if too small, don't report yet
		if (subg.VertexLabel().size() != subg_full.VertexLabel().size())
			return;

		std::cout << "Full subgraph seen, checking edges" << std::endl;

		// Check graph is by comparing edges
		if (subg.Edges() != subg_full.Edges())
			return;

		std::cout << "    edges equal" << std::endl;

		// Must be unique
		assert(0 == *std::max_element(count.begin(), count.end()));
		for (std::map<unsigned int, std::vector<std::vector<unsigned int> > >::const_iterator
			li = L.begin () ; li != L.end () ; ++li) {
			count[li->first] = li->second.size();
		}

		std::cout << "count:" << std::endl
			<< gmm::row_vector(count) << std::endl;
	}

public:
	MexIsGeoSubgraph (const std::vector<FriGeo::Graph>& G,
		const FriGeo::Graph& subg, double vertexTolerance)
		: FriGeo::FreqGeo (G, subg, 1, vertexTolerance),
			subg_full(subg)
	{
		count.resize(G.size());
		std::fill(count.begin(), count.end(), 0);
	}

	/* [count]
	 *
	 * count: (n,1) integer frequency count of the found graphs.
	 */
	void
	writeMexLHS (int nlhs, mxArray* plhs[])
	{
		if (nlhs >= 1) {
			uint32_T* countP = NULL;
			plhs[0] = mxCreateNumericMatrix (count.size (), 1, mxUINT32_CLASS, 0);
			assert (plhs[0] != NULL);
			countP = (uint32_T *) mxGetPr (plhs[0]);
			std::copy(count.begin(), count.end(), countP);
		}
	}
};


/* [occurence] = mexisgeosubgraph (G, subg, tolerance);
 */
void
mexFunction (int nlhs, mxArray* plhs[], int nrhs, const mxArray* prhs[]) {
	if (nrhs < 2 || nrhs > 3) {
		mexErrMsgTxt ("Wrong number of right hand side parameters (must be 2 or 3).\n");

		return;
	}

	if (nlhs < 1 || nlhs > 1) {
		mexErrMsgTxt ("Wrong number of left hand side parameters (must be 1).\n");

		return;
	}

	if (mxIsCell (prhs[0]) == 0) {
		mexErrMsgTxt ("G must be a cellarray of graph structures.\n");

		return;
	}

	const mxArray* graphs = prhs[0];
	unsigned int gcount = mxGetN (graphs);

	const mxArray** garr = (const mxArray**) mxGetPr (graphs);
	std::vector<FriGeo::Graph> G;

	for (unsigned int n = 0 ; n < gcount ; ++n) {
		const mxArray* gcur = garr[n];

		FriGeo::Graph g;
		if (FriGeo::MexHelp::readGraph (gcur, g) == false)
			return;

		G.push_back (g);
	}

	FriGeo::Graph subg;
	if (FriGeo::MexHelp::readGraph (prhs[1], subg) == false)
		return;

	double tolerance = 1e-1;
	if (nrhs >= 3)
		tolerance = mxGetScalar (prhs[2]);

	if (tolerance < 0.0) {
		mexErrMsgTxt ("The given matching tolerance is negative.\n");

		return;
	}

	MexIsGeoSubgraph fg (G, subg, tolerance);
	fg.Mine ();
	fg.writeMexLHS (nlhs, plhs);
}


